import sys
import os
import inspect
import logging
from logging import handlers


# ---------------------------------------------------------------------------------------
# Creat new logger levels

DEBUG_LEVELV_NUM = 6
logging.addLevelName(DEBUG_LEVELV_NUM, "DEBUGV")
def debugv(self, message, *args, **kws):
    # Yes, logger takes its '*args' as 'args'.
    if self.isEnabledFor(DEBUG_LEVELV_NUM):
        self._log(DEBUG_LEVELV_NUM, message, args, **kws)

logging.Logger.debugv = debugv

DEBUG_LEVELX_NUM = 8
logging.addLevelName(DEBUG_LEVELX_NUM, "DEBUGX")
def debugx(self, message, *args, **kws):
    # Yes, logger takes its '*args' as 'args'.
    if self.isEnabledFor(DEBUG_LEVELX_NUM):
        self._log(DEBUG_LEVELX_NUM, message, args, **kws)

logging.Logger.debugx = debugx

DEBUG_LEVELW_NUM = 12
logging.addLevelName(DEBUG_LEVELW_NUM, "INFOW")
def infow(self, message, *args, **kws):
    # Yes, logger takes its '*args' as 'args'.
    if self.isEnabledFor(DEBUG_LEVELW_NUM):
        self._log(DEBUG_LEVELW_NUM, message, args, **kws)

logging.Logger.infow = infow
# ---------------------------------------------------------------------------------------


log = None
logcsv = None


def init_logger(logger=None, name='logger_util'):
    global log
    if logger is not None:
        log = logger
    else:
        import logging
        log = logging.getLogger(name)

def init_csvlogger(logger=None, name='logger_util'):
    global logcsv
    if logger is not None:
        logcsv = logger
    else:
        import logging
        logcsv = logging.getLogger(name)

global logs_fully_setup
logs_fully_setup = None

def get_logger(
    name=None, applog="pegasus1.log", basiclog="pegasus_B.log", advlog="pegasus_ADV.log", level=logging.DEBUG
):
    default = "__app__"
    # formatter = logging.Formatter('%(thread)d : %(levelname)s: %(asctime)s %(funcName)s(%(lineno)d) -- %(message)s',
    formatter = logging.Formatter(
        "%(threadName)s: %(levelname)s: %(asctime)s %(funcName)s(%(lineno)d) -- %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    log_map = {"radio_switch": applog, "__app__": applog, "__basic_log__": basiclog, "__advance_log__": advlog}
    if name:
        logger = logging.getLogger(name)
    else:
        logger = logging.getLogger(default)
    # fh = logging.FileHandler(log_map[name])
    # fh.setFormatter(formatter)
    # logger.addHandler(fh)
    consoleHandler = logging.StreamHandler(sys.stdout)
    logger.addHandler(consoleHandler)
    logger.setLevel(level)
    return logger


def setup_sockt_logger(thelogger, logger_ip):
    socketHandler = logging.handlers.SocketHandler(logger_ip, logging.handlers.DEFAULT_TCP_LOGGING_PORT)
    # don't bother with a formatter, since a socket handler sends the event as
    # an unformatted pickle
    thelogger.addHandler(socketHandler)


def setup_logger(name, fname, level=logging.INFO, mode=''):
    home = os.path.expanduser("~")

    currentFile = __file__  # May be 'my_script', or './my_script' or
    if fname == "":
        fname = os.path.splitext(os.path.basename(currentFile))[0]

    logfname = fname
    # '/home/user/test/my_script.py' depending on exactly how
    # the script was run/loaded.
    realPath = os.path.realpath(currentFile)  # /home/user/test/my_script.py
    dirPath = os.path.dirname(realPath)  # /home/user/test
    dirName = os.path.basename(dirPath)  # test

    if "core" in mode:
        log_dir = "/tmp/x_task_logs/logs/"
        log_bin_dir = "/tmp/x_task_logs/bin/"
    else:
        log_dir = dirPath + "/x_task_logs/logs/"
        log_bin_dir = dirPath + "/x_task_logs/bin/"

    if not os.path.exists(log_dir):
        os.makedirs(log_dir)
    if not os.path.exists(log_bin_dir):
        os.makedirs(log_bin_dir)

    print("LOGGING TO -> {}".format(log_dir, logfname))

    """Function setup as many loggers as you want"""
    formatter = logging.Formatter("%(message)s")
    handler = logging.FileHandler(log_dir + logfname + ".log")
    handler.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.setLevel(level)
    # logger.addHandler(handler)

    return logger


def setuplogs(fname="", name="__app__", level=logging.INFO, socket_logger_ip=None, force_setup=False, mode=''):
    global logs_fully_setup, log_dir, log_bin_dir
    if not force_setup and logs_fully_setup:
        return logging.getLogger(name)

    home = os.path.expanduser("~")

    currentFile = __file__  # May be 'my_script', or './my_script' or
    if fname == "":
        fname = os.path.splitext(os.path.basename(currentFile))[0]

    logfname = fname
    # '/home/user/test/my_script.py' depending on exactly how
    # the script was run/loaded.
    realPath = os.path.realpath(currentFile)  # /home/user/test/my_script.py
    dirPath = os.path.dirname(realPath)  # /home/user/test
    dirName = os.path.basename(dirPath)  # test

    log_dir = './logs'
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)

    print("Console LOGGING enabled")
    thelogger = get_logger(name=name, applog=log_dir + logfname + ".log", level=level)

    # if socket_logger_ip is not None:
    #     setup_sockt_logger(thelogger, socket_logger_ip)

    logs_fully_setup = True
    #thelogger.info(f"{name} started, logging to file: " + logfname + ", path: " + log_dir)
    return thelogger




import os
from os.path import exists
import pickle
import time
import timeit
import scipy.misc as sc
import math
import numpy as np
from scipy.stats import binom
import scipy.stats as st
from scipy import special as scm
import scipy.stats as st
from scipy.stats import binom
from scipy.special import factorial, comb, perm
import gmpy2
from .vsa_stats import *
from .helpers import deserialise_object


class IntegerParts(object):
    def __init__(self, elements, counts):
        self.elments = elements
        self.counts = counts

    @property
    def len(self):
        return np.sum(self.counts)


class SnnHd(object):
    def __init__(self, B, m):
        self.B = B
        self.m = m
        self.B_pow_m1 = gmpy2.mpz(B) ** gmpy2.mpz(m - 1)

        self.Pn = {}
        self.AnW = {}
        self.RnW = {}
        self.AnD = {}
        self.RnD = {}

    def calc_prob(self, Pn, C, D):
        hash_C = C.data.tobytes()
        if D == 0:
            W = 1
            C[-1] -= 1  # ** modifying C
            if hash_C in self.AnW:
                An = self.AnW[hash_C]
                Rn = self.RnW[hash_C]
            else:
                An = num_arrangements(C)
                self.AnW[hash_C] = An
                Rn = num_repeats(self.B, C)
                self.RnW[hash_C] = Rn
            C[-1] += 1  # ** fixup C
        else:
            W = C[-1] + 1
            C[-2] -= 1  # ** modifying C
            if hash_C in self.AnD:
                An = self.AnD[hash_C]
                Rn = self.RnD[hash_C]
            else:
                An = num_arrangements(C)
                self.AnD[hash_C] = An
                Rn = num_repeats(self.B, C)
                self.RnD[hash_C] = Rn
            C[-2] += 1  # ** fixup C

        P = Pn * An * Rn / (gmpy2.mpz(W) * self.B_pow_m1)

        return float(P)


def num_patterns(E, C, m):
    F = np.array([gmpy2.factorial(gmpy2.mpz(x)) for x in E])  # factorial(E)
    Fp = np.power(F, C)  #  np.power(F, C)
    Pn = gmpy2.factorial(m-1) / np.product(Fp) # factorial(m - 1) / np.product(Fp)
    return Pn


def num_arrangements(C):
    F = np.array([gmpy2.factorial(gmpy2.mpz(x)) for x in C])  # factorial(C)
    An = gmpy2.factorial(gmpy2.mpz(sum(C))) / np.product(F)
    return An


def num_repeats(B, C):
    Rn = gmpy2.comb(gmpy2.mpz(B) - 1, gmpy2.mpz(sum(C)))
    return Rn


def accel_asc(n, l):
    a = np.zeros(n + 2)
    k = 1
    a[0] = 0
    a[1] = n
    while k != 0:
        x = a[k - 1] + 1
        y = a[k] - 1
        k -= 1
        while x <= y and k < l - 1:
            a[k] = x
            y -= x
            k += 1
        a[k] = x + y
        a[k + 1] = 0     # We want a trailing zero
        yield a[:k + 2]  # We want a trailing zero


def sparse_error(slots, bits, target_hd, vocab_size):
    return perror(slots, target_hd, 1.0/bits, 1e10, vocab_size)


def subvec_mean(m, B, M, use_disk_files=False, diskfile_path='./integer_partitions'):
    P_sum = 0
    if use_disk_files and exists(f"{diskfile_path}/ip{m-1:04d}.bin"):
        return get_snn_hsim_from_file(m, B, M, diskfile_path)

    mysnn_calc = SnnHd(B, m)
    for value in accel_asc(m - 1, B):
        a = value[::-1]
        # need zero on the start of the elements and counts
        elements, counts = np.asarray(np.unique(a, return_counts=True))
        counts[0] = 0

        if True:  # or zpart.len <= B:
            E = elements
            C = counts
            D = 0
            myPn = num_patterns(E, C, m)
            P = mysnn_calc.calc_prob(myPn, C, D)
            P_sum += P
            if P >= 1.0:
                print(m, P, P_sum, 'W')

            # If combi can be a shared win (E[-1] - E[-2] == 1) need to add in this probability also.
            l = len(E) - 1
            if l >= 1 and E[-1] - E[-2] == 1:
                D = 1
                if E[-1] == 1:
                    C[0] = 1  # We are modifying C but it is not used elsewhere after modification.
                P = mysnn_calc.calc_prob(myPn, C, D)
                P_sum += P
                if P >= 1.0:
                    print(m, P, P_sum, 'D')

    return float(P_sum * M)


def get_snn_hsim_from_file(m, B, M, diskfile_path='./integer_partitions'):
    P_sum = 0
    ipart_name = f"{diskfile_path}/ip{m-1:04d}.bin"
    iparts = deserialise_object(ipart_name, None)
    mysnn_calc = SnnHd(B, m)
    for zpart in iparts:
        if zpart.len <= B:
            E = zpart.elments
            C = zpart.counts
            D = 0
            myPn = num_patterns(E, C, m)
            P = mysnn_calc.calc_prob(myPn, C, D)
            P_sum += P
            if P >= 1.0:
                print(m, P, P_sum, 'W')

            # If combi can be a shared win (E[-1] - E[-2] == 1) need to add in this probability also.
            l = len(E) - 1
            if l >= 1 and E[-1] - E[-2] == 1:
                D = 1
                if E[-1] == 1:
                    C[0] = 1  # We are modifying C but it is not used elsewhere after modification.
                P = mysnn_calc.calc_prob(myPn, C, D)
                P_sum += P
                if P >= 1.0:
                    print(m, P, P_sum, 'D')

    return float(P_sum * M)



import os
from os.path import exists
import pickle
import datetime
from enum import IntEnum
import numpy as np
import vsapy.vsapy as vsa
from vsapy.vsatype import *
from vsapy.helpers import serialise_object


class TimeStamp(object):
    @staticmethod
    def get_creation_data_time_stamp():
        return 'Timestamp: {:%Y-%m-%d %H:%M:%S.%f}'.format(datetime.datetime.now())

    @staticmethod
    def compare_time_stamps(t1, t2):
        t1obj = datetime.datetime.strptime(t1, 'Timestamp: %Y-%m-%d %H:%M:%S.%f')
        t2obj = datetime.datetime.strptime(t2, 'Timestamp: %Y-%m-%d %H:%M:%S.%f')
        return t1obj.time() == t2obj.time()


def createSymbolVectors(symbols, *args, creation_data_time_stamp=None, **kwargs):
    if creation_data_time_stamp is None:
        creation_data_time_stamp = TimeStamp.get_creation_data_time_stamp()
    # A dictionary is slightly faster than Graham's list look up but only marginally....
    sym = {"creation_data_time_stamp": creation_data_time_stamp}

    for a in symbols:
        sym[a] = vsa.randvec(*args, **kwargs)

    return sym


def create_base_vecs(start, end, veclen, ascii_names=True,
                     creation_data_time_stamp=None, vsa_type=VsaType.BSC, **kwargs):

    if creation_data_time_stamp is None:
        time_stamp = TimeStamp.get_creation_data_time_stamp()
    else:
        time_stamp = creation_data_time_stamp

    base_vecs = {"creation_data_time_stamp": time_stamp}  # Dictionary(Of String, BitArray)
    if ascii_names:
        start = ord(start)
        end = ord(end)
    for x in range(start, end + 1):  # inclusive of end
        if ascii_names:
            c = chr(x)
        else:
            c = x
        base_vecs.update({c: vsa.randvec(veclen, vsa_type=vsa_type, **kwargs)})
        list(base_vecs.items())

    return base_vecs


class PackedVec(object):
    '''
    Used to reduce memory storage overhead by keeping vectors packed.
    '''
    def __init__(self, v):
        self.vsa_type = v.vsa_type
        self.myvec = v

    @property
    def myvec(self):
        # performance for Laiho/X is slightly improved by not searching the subclass chain
        if self.vsa_type == VsaType.Laiho or self.vsa_type == VsaType.LaihoX or self.vsa_type == VsaType.HRR:
            return self.__myvec
        else:
            return VsaBase.get_subclass(self.vsa_type).unpackbits(self.__myvec)

    @myvec.setter
    def myvec(self, v):
        # performance for Laiho/X is slightly improved by not searching the subclass chain
        if isinstance(v, (vsa.Laiho, vsa.HRR)):
            self.__myvec = v
        else:
            self.__myvec = v.packbits(v)

    @property
    def packed(self):
        return self.__myvec


class BagVec(PackedVec):
    def __init__(self, veclist, vec_cnt=-1, myvec=None):
        """

        :param veclist: list of vectors to be majority-summed.
        :param vec_cnt: If the vector is already summed the number of vectors in the sum. This allows a mix of
                        normalised and un-normalised vectors to be passed in veclist. For example,
                        veclist = [[2,1,3,5,0], [1,0,1,1,0,1]] then vec_cnt might = 7
                        because 1st vec could be an un-normallised sum of 6 vecs and the second vec is normalised.
        """
        if myvec is None:
            rawvec, vec_cnt, myvec = BagVec.bundle(veclist, vec_cnt)
        else:
            assert vec_cnt >= 1, "you must specify parameter 'vec_cnt': number of vectors in this vector."

        super(BagVec, self).__init__(myvec)
        self.vec_cnt = vec_cnt

    @property
    def vec_cnt(self):
        return self.__vec_cnt

    @vec_cnt.setter
    def vec_cnt(self, vcount):
        self.__vec_cnt = vcount

    @staticmethod
    def bundle(veclist, vec_cnt=-1):
        if isinstance(veclist, np.ndarray):
            if len(veclist.shape) == 1:   # i.e. only 1D array
                assert vec_cnt >= 1, "you must specify parameter 'vec_cnt': number of vectors in this vector."
                rawvec = veclist  # veclist is a single [a1, a2, a3, ...] numpy column vector
                return rawvec, vec_cnt,  vsa.normalize(veclist, vec_cnt, )
        elif isinstance(veclist, list):
            if not isinstance(veclist[0], (list, np.ndarray)):  # i.e. only 1D array
                assert vec_cnt >= 1, "you must specify parameter 'vec_cnt': number of vectors in this vector."
                rawvec = veclist  # veclist is a single [a1, a2, a3, ...] python list vector
                return rawvec, vec_cnt,  vsa.normalize(veclist, vec_cnt, )
        else:
            raise ValueError(" 'veclist' is not an array type.")

        # veclist contains a list of vectors t be added.
        sumv = veclist[0].sum(veclist)
        if isinstance(sumv, vsa.Laiho):
            rawvec = None
            norm_vec = sumv
            vec_cnt = len(veclist)
        else:
            rawvec = sumv
            if vec_cnt >= 1:
                # This enables passing in one or more un-normalized vectors in veclist. If vec_cnt >= 1 we assume that
                # vec_cnt accounts for all un-normalised vectors in the list
                norm_vec = vsa.normalize(sumv, vec_cnt, )
            else:
                norm_vec = vsa.normalize(rawvec, len(veclist), )
                vec_cnt = len(veclist)

        return rawvec, vec_cnt, norm_vec


class RawVec(BagVec):
    def __init__(self, veclist, vec_cnt=-1):
        rawvec, vec_cnt, myvec = BagVec.bundle(veclist, vec_cnt)
        super(RawVec, self).__init__(None, vec_cnt, myvec=myvec)

        if isinstance(veclist[0], (np.ndarray, list)) and isinstance(veclist[0], vsa.Laiho):
            self.__rawvec = self.myvec
        else:
            self.__rawvec = rawvec

    @property
    def rawvec(self):
        return self.__rawvec

    @rawvec.setter
    def rawvec(self, rawvec):
        self.__rawvec = rawvec


def serialise_vec_hierarchy(chunk_hierarchy, pathfn):
    bare_hamlet = BareChunk(chunk_hierarchy)
    serialise_object(bare_hamlet, pathfn)
    return


class BareChunk(RawVec):
    def __init__(self, cnk):
        super(BareChunk, self).__init__(cnk.rawvec, cnk.vec_cnt)
        self.vsa_type = cnk.vsa_type
        self.creation_data_time_stamp = cnk.creation_data_time_stamp
        self.aname = cnk.aname
        self.chunklist = [BareChunk(c) for c in cnk.chunklist] if not cnk.isTerminalNode else []
        self.__terminal_node = cnk.isTerminalNode

    @property
    def isTerminalNode(self):
        return self.__terminal_node

    def get_acts(self, name_prefix=''):
        if name_prefix != '':
            acts = []
            for act in self.chunklist:
                act.aname = f'{name_prefix}@{act.aname}'
                acts.append(act)
            return acts
        else:
            return [act for act in self.chunklist]

    def flattenchunkheirachy(self, allchunks, skip_worker_chunks=False):
        if self.isTerminalNode:
            if skip_worker_chunks:
                return
            else:
                allchunks.append(self)
            return
        else:
            allchunks.append(self)
            for c in self.chunklist:
                c.flattenchunkheirachy(allchunks, skip_worker_chunks)

    @staticmethod
    def get_level_number(aname):
        return aname[aname.find("$")+1:aname.find("@")]

    @staticmethod
    def add_level_labels(cnk, current_level, doc_prefix, label=''):
        if current_level == 0 or label == '':
            label = '00'
        else:
            label = f"{label}-{current_level:02d}"

        cnk.aname = f"{doc_prefix}${label}@{cnk.aname}"
        if not cnk.isTerminalNode:
            current_level = 1
            for c in cnk.chunklist:
                BareChunk.add_level_labels(c, current_level, doc_prefix, label)
                current_level += 1

        return

    @staticmethod
    def get_levels_as_list(cnk, levels, current_level_request, cnk_list):
        """
        Note this method retrieves chunks directly from a stored heirarchical chunk tree - NOT by using unbinding.
        :param cnk: top level chunk from which to start the retrieval
        :param levels:  pass a list of the levels you want to retrieve, in this case [0, 1] = the whole play and each
                act. Passing [2] would retrieve only the scene data.
        :param current_level_request: recursive leval tracker passing zero to start should always work
        :param cnk_list: flat list of retrieved chunks. Passed as parameter because of recursion.
        """
        ll = levels[:]
        ll.sort()  # Levels need to be in order for retrieval to work.
        if current_level_request == ll[0]:
            cnk_list.append(cnk)
            ll = ll[1:]  # pinch off the first element

        # descend cnk hierarchy until we are at level zero
        if len(ll) > 0:
            current_level_request += 1
            for c in cnk.chunklist:
                BareChunk.get_levels_as_list(c, ll, current_level_request, cnk_list)

        return

class BareChunkRunDetail(object):
    '''
    Saves a chunk hierarchy with details of how it was built, e.g.
            input
            encoding scheme for leaf nodes:
                    chain (unique vectors for each word),
                    pvec (syntatic matching)
                    WV-T = True, WV-F = False - whether word2vec symantic vectors where used for leaf nodes.
                    CW-T = True, CW-F = False - hierarchy descends to word level for leaf nodes

    '''

    def __init__(self, chunk_hierarchy, run_detail, play_name, short_info, long_info):
        self.play_name = play_name
        self.run_detail = run_detail
        self.short_info = short_info
        self.long_info = long_info
        self.top_chunk = chunk_hierarchy



import numpy as np
from scipy.stats import binom


def perror(nobits, target_hd, alternate_hd, num_trials, vocab_size):
    P_hd = binom.pmf(range(0, nobits), nobits, target_hd)
    #P_n = binom.pmf(range(0, nobits), nobits, alternate_hd)
    P_M= (binom.cdf(range(0, nobits), nobits, alternate_hd)) ** vocab_size
    P_match = P_hd*P_M
    #HD0_all = (P_hd * num_trials).astype(int)
    all_wins = (P_match*num_trials).astype(int)
    P_win = np.sum(all_wins)/num_trials
    P_error = 1 - P_win

    return P_error



import math
from decimal import Decimal
import numpy as np
from scipy import special as scm
import gmpy2
from .vsa_stats import *


def get_perror(nobits, target_hd, alternate_hd, num_trials, vocab_size):
    return perror(nobits, 1.0 - target_hd, 1.0-alternate_hd, num_trials, vocab_size)


def subvec_mean(num_vecs):
    """
    :param num_vecs: This is the number of vectors to be added via majority_sum
    :return: The normalised hamming distance similarity between an individual vector of the sum and the majority_sum.
             0 = exact match. If the vectors being added are 1000 bits long then a result of 0.25 means that
             of the 1000 bits only 250 bits will be different than those contained in each individual vector.
             Or putting it the other way around 75% of the bits in each individual vector in the sum will match to the
             bits in the resultant majority_sum vector.
    """

    if num_vecs == 1:
        return float(1.0)

    # Majority sum needs an odd number of vectors, an additional random vector is used when the sum contains
    # and even number of vectors, hence if we have an even number here it is equivalent to calculating on the
    # num_vecs + 1
    if num_vecs % 2 > 0:  # If odd
        num_vecs -= 1

    P = 0  # Cumulative permutation sum
    for j in range(num_vecs // 2, num_vecs + 1):
        P = P + gmpy2.comb(num_vecs, j)

    for _ in range(num_vecs):
        P /= 2

    return float(P)


def sum_combis(N, k):
    combis = gmpy2.mpz(0)
    for j in range(k, N + 1):
        combis += gmpy2.comb(N, j)
    return float(combis / 2 ** N)


def get_probability_that_bit_is_set(total_subvecs_in_this_sum, num_shared_vecs,
                                    majority_bit_cnt_in_this_shared_vec_instance):
    """

    :param total_subvecs_in_this_sum: Number of sub vecs

    :param num_shared_vecs:  Number of vecs common to some other vector

    :param majority_bit_cnt_in_this_shared_vec_instance:
        Imagine that there are two shared vectors, then this parameter indicates the number of bits that are in the
        majority for this instance of the shared vecs, for example, when num_shared_bits = 2, all possibilities are
            00
            01
            10
            11

        When the '01' instance is being processed this param = 1, when 00 or 11 is being processed this param = 2

    :return: Probability that the bit in Z/Z' will have the same sense as the majority in the shared vecs.
    """
    wining_line = int(total_subvecs_in_this_sum / 2) + 1  # winning count for Majority_Sum
    other_vecs = total_subvecs_in_this_sum - num_shared_vecs

    if majority_bit_cnt_in_this_shared_vec_instance >= wining_line:
        # If shared vecs have put us over the winning line then other_bits='Don't Care'
        return 1.0
    else:
        if float(majority_bit_cnt_in_this_shared_vec_instance) == float(num_shared_vecs) / 2.0:
            # Then all winnig bits come Non-shared-bit-columns
            # If there is a tie in the shared_bits then winner comes solely from other side
            # i.e. the winner is the majority occurrence chosen from (other_vecs ONLY)
            p = 0.5  # when choosing a majority of 1, e.g 3c2 or 11c6 the probability is always p=0.5
        else:
            extra_bits_needed = wining_line - majority_bit_cnt_in_this_shared_vec_instance
            p = sum_combis(other_vecs, extra_bits_needed)

    return p  # Probability that the bit in Z/Z' will have the same sense as the majority in the shared vecs.


def get_one_sided_shared_vec_probability_profile(total_subvecs_in_this_sum, num_shared_vecs):
    """

    This has O(N) time complexity.

    :param total_subvecs_in_this_sum: number of sub_vecs in the sum (i.e. the target_vector=Z)
    :param num_shared_vecs: number of sub_vecs that might be shared in some other Z'
    :return: theoretical hamming similarity
    """

    """--------------------------------------------------------------------------------------------------------------
    Map the number of dominant bits (e.g. 1s) for all possible occurences of num_shared_vecs against frequency of 
    occurence, for example, when num_shared_bits = 2, all possibilities are
        00
        01
        10
        11
                                            10  
                                 00         01         11
    Thus shared_vec_combi_map = [(2, 0.25), (1, 0.5),  (2, 0.25)]
                                     p=25%      p=50%      p=25%

    """
    DEBUG_PREPROC = False
    shared_vec_combi_map = []
    shared_vec_combi_map_check = []
    for i in range((num_shared_vecs+1) // 2, num_shared_vecs+1):
        if DEBUG_PREPROC:
            shared_vec_combi_map_check.append((i, scm.comb(num_shared_vecs, i)))

        shared_vec_combi_map.append((i, scm.comb(num_shared_vecs, i)/2**num_shared_vecs))
    """--------------------------------------------------------------------------------------------------------------"""

    prob_map = []
    for shared_ones_cnt, prob_of_this_shared_bit_instance_occurring in shared_vec_combi_map:
        n_pwr = num_shared_vecs // shared_ones_cnt - 1
        divisor = 2 ** n_pwr
        prob_map.append((get_probability_that_bit_is_set(total_subvecs_in_this_sum, num_shared_vecs, shared_ones_cnt),
                         2.0/float(divisor) * prob_of_this_shared_bit_instance_occurring, shared_ones_cnt))

    """--------------------------------------------------------------------------------------------------------------
    Continuing the example, get_probability_that_bit_is_set() calculates the probability that the bit in Z/Z' 
    will have the same sense as the majority in the shared vecs, for num_shared_bits = 2,

                                 10  
                        00       01          11
    Thus prob_map = [(1, 0.25), (0.5, 0.5),  (1, 0.25)]
            occurs:       25%          50%         25%
       contributes: 100%         50%        100%        example, depends on total_sub_vecs.. and num_shared_vecs

    """
    return prob_map


def get_shared_hamming_similarity(total_no_subvecs_in_V1_sum, num_shared_vecs, total_no_subvecs_in_V2_sum):
    """

    This has O(N) time complexity and is much quicker

    :param total_no_subvecs_in_V1_sum:
    :param num_shared_vecs:
    :param total_no_subvecs_in_V2_sum:
    :return: probability that Z=Z' which is also the hamming similarity
    """

    if num_shared_vecs == 0:
        return 0.5  # If there are no shared vecs then by definition the vectors are orthogonal, i.e. HDSim = 0.5

    assert num_shared_vecs <= min(total_no_subvecs_in_V1_sum, total_no_subvecs_in_V2_sum), \
        "parameter 2, num_shared_vecs can not be greater than total vecs contained in a sum vector."
    v1_probability_map = get_one_sided_shared_vec_probability_profile(total_no_subvecs_in_V1_sum, num_shared_vecs)
    v2_probability_map = get_one_sided_shared_vec_probability_profile(total_no_subvecs_in_V2_sum, num_shared_vecs)

    sum_prob1 = 0
    for i in range(len(v1_probability_map)):
        # The probability of each shared_bit_instance, that is this particular instance combination of 1s and 0's,
        # should always be equal since it depends solely on the number of vecs common to both vectors
        assert v1_probability_map[i][1] == v2_probability_map[i][1], "Shared_probabilities_do_not_match"

        pv11 = v1_probability_map[i][0]  # probability that the bit is set, i.e. Z = 1, in vector 1
        pv22 = v2_probability_map[i][0]  # probability that the bit is set, i.e. Z'= 1, in vector 2


        # The probability that Z=Z' works both ways, that is
        # We need (prob_V1_occurring) * (prob_V2_occurring) + (prob_V1_NOT_occurring) * (prob_V2_NOT_occurring)
        prob1 = pv11 * pv22 + (1 - pv11) * (1 - pv22)

        sum_prob1 += v1_probability_map[i][1] * prob1  # prob of this shared_bit_config * prob of Z=Z'

    return sum_prob1



import math
from functools import reduce
from .vsatype import *


class HRR(VsaBase):
    vsatype = VsaType.HRR

    @classmethod
    def default_numpy_type(cls):
        """
        :return: the default numpy datatype for this type of VSA.
        """
        return 'float'

    @classmethod
    def unpackbits(cls, v):
        # HRR vecs are not packed
        return v

    @classmethod
    def packbits(cls, v):
        # HRR vecs are not packed
        return v

    @classmethod
    def random_threshold(cls, *args, stdev_count=4.4, **kwargs):
        """
        Should return a normalised value of the similarity match that would be expected when comparing
        random/orthorgonal vectors.

        :param args: A sample vector the dimensions of which should be used to calculate threshold.
        :param stdev_count:
        :return: threshold value
        :rtype: float
        """
        if len(args) > 0:
            if isinstance(args[0], HRR):
                D = len(args[0])
            elif isinstance(args[0], int):
                D = args[0]
            else:
                raise ValueError("Expected a HRR vector or an int specifying vector dimension.")
        else:
            raise ValueError("You must supply a sample vector, or a value for dimensionality.")

        p = 0.0  # For cosine Mean Probability of a match between random vectors = zero
        var_rv = float(1/D)  # Varience (un-normalised)
        std_rv = math.sqrt(var_rv)  # Stdev (un-normalised)
        hdrv = 0.0 + stdev_count * std_rv  # Un-normalised hsim of two randomvectors adjusted by 'n' stdevs
        return hdrv  # No need to normalise, it already is.

    @classmethod
    def randvec(cls, dims, word_size=8, vsa_type=VsaType.HRR):
        """
        :param dims: integer or tuple, specifies shape of required array, last element is no bits per vector.
        :param word_size: numpy's word size parameter, e.g. for BSCs wordsize=8 becomes 'uint8'.
        :param vsa_type: type of VSA subclass to create from VsaType class.
        :return: a matrix of vectors of shape 'dims'.
        """
        return VsaBase(np.random.uniform(-1.0, 1.0, dims), vsa_type=VsaType.HRR)

    @classmethod
    def bind(cls, a, b):
        """
        Comutative binding operator
        :param a: VSA vec
        :param b: VSA vec
        :return: vector associating/coupling a to b that is dissimilar to both a and b.
                 In most cases bind(a, b) is analogues to multiplication, e.g. bind(3,4)=>12.
                 If we know one of the operands we can recover the other using unbind(a,b) e.g unbind(3,12)=>4
        """
        assert a.vsa_type == b.vsa_type, "Mismatch vsa_types"
        if a.vsa_type == VsaType.HRR and a.vsa_type == b.vsa_type:
            return cls.cconv(a, b)
        raise ValueError("Mismatch vsa_types")

    @classmethod
    def unbind(cls, a, b):
        """
        Comutative unbinding operator. Decouples a from b and vice-versa. The result
        :param a: VSA vec
        :param b: VSA vec
        :return: reverses a bind operation. If z = bind(x, y) then x = unbind(y, z) and y = unbind(x, z).
                 The return is orthogonal to x nd y if x and y have not been previously associated with bind(x, y).
        """
        assert a.vsa_type == b.vsa_type, "Mismatch vsa_types"
        if b.vsa_type == VsaType.HRR:
            return cls.ccorr(a, b)
        raise ValueError("Mismatch vsa_types")

    @classmethod
    def normalize(cls, sv, seqlength=None, rv=None):
        """
        Normalize the VSA vector
        :param a: input VSA vector
        :param seqlength: not used
        :param rv: Optional random vector, used for splitting ties on binary and ternary VSA vectors.
        :return: new VSA vector
        """
        assert sv.vsa_type == VsaType.HRR, "Mismatch vsa_types"
        if sv.vsa_type == VsaType.HRR:
            """
            Normalize a vector to length 1.
            :param a: Vector
            :return: a / len(a)
            """
            v = np.asarray(sv)
            return VsaBase(v / np.sum(v ** 2.0) ** 0.5, vsa_type=VsaType.HRR)
        raise ValueError("Mismatch vsa_types")

    @classmethod
    def hdist(cls, a, b):
        """
        Cosine is used for hsim() for real number vectors
        :param a: vsa vector
        :param b: vsa vector
        :return: cosine similarity between a and b. 0.0=exact match.
        """
        return cls.cosine(a, b)

    @classmethod
    def hsim(cls, a, b):
        """
        Cosine is used for hsim() for real number vectors
        :param a: vsa vector
        :param b: vsa vector
        :return: cosine similarity between a and b. 1.0=exact match.
        """
        assert a.vsa_type == a.vsa_type, "Mismatch vsa_types"
        if b.vsa_type == VsaType.HRR:
            return abs(1.0 - cls.cosine(a, b))
        raise ValueError("Mismatch vsa_types")

    @classmethod
    def cconv(cls, a, b):
        """
        Computes the circular convolution of the (real-valued) vectors a and b.
        """
        assert a.vsa_type == b.vsa_type, "Mismatch vsa_types"
        return VsaBase(np.fft.ifft(np.fft.fft(a) * np.fft.fft(b)).real, vsa_type=VsaType.HRR)

    @classmethod
    def ccorr(cls, a, b):
        """
        Computes the circular correlation (inverse convolution) of the real-valued
        vector a with b.
        """
        assert a.vsa_type == b.vsa_type, "Mismatch vsa_types"
        return cls.cconv(np.roll(a[::-1], 1), b)

    @classmethod
    def ordConv(cls, a, b, p1, p2):
        """
        Performs ordered (non-commutative) circular convolution on the vectors a and
        b by first permuting them according to the index vectors p1 and p2.
        """
        assert all(v == a for v in [b, p1, p1]), "Mismatch vsa_types"
        return cls.cconv(a[p1], b[p2])

    @classmethod
    def convBind(cls, p1, p2, l):
        """
        Given a list of vectors, iteratively convolves them into a single vector
        (i.e., "binds" them together).
        """
        return reduce(lambda a, b: cls.normalize(cls.ordConv(a, b, p1, p2)), l)

    @classmethod
    def addBind(cls, p1, p2, l):
        """
        Given a list of vectors, binds them together by iteratively convolving them
        with place vectors and adding them up.
        """
        return cls.normalize(reduce(lambda a, b: cls.cconv(a, p1) + cls.cconv(b, p2), l))



from enum import IntEnum
import numpy as np
import scipy.spatial as sp


class VsaType(IntEnum):
    BSC = 1  # i.i.d Binary spatter code
    Tern = 2  # i.i.d Ternary spatter code (+1, -1)
    TernZero = 3  # Ternary spatter code but allows don't care (+1, 0, -1) where 0 means tie occured in bundling
    HRR = 4  # Holographic Reduced Representation
    Laiho = 5  # Laiho full implementation
    LaihoX = 6  # Laiho simplified - the first nonzero element is taken in bundling operations


class VsaBase(np.ndarray):
    vsatype = None

    class NoAccess(Exception): pass
    class Unknown(Exception): pass

    def __new__(cls, input_array, vsa_type, dtype=None, bits_per_slot=None):
        """
        Create instance of appropriate VsaBase subclass using vsa_type.
        :param input_array:
        :param vsa_type: subclass of VsaBase to be created from VsaType class
        :param dtype: numpy dtype
        """
        subclass = cls.get_subclass(vsa_type)
        if subclass:
            # Using "object" base class method avoids recursion here.
            obj = np.asarray(input_array).view(subclass)
            # add the new attribute to the created instance
            obj.vsa_type = vsa_type
            if vsa_type == VsaType.Laiho or vsa_type == VsaType.LaihoX:
                assert bits_per_slot is not None, "you must supply value for bits_per_slot"
                obj.slots = input_array.shape[-1]
                obj.bits_per_slot = bits_per_slot

            # Finally, we must return the newly created object, cast to the required numpy dtype:
            if dtype:
                return obj.astype(dtype)
            else:
                return obj.astype(obj.default_numpy_type())

        else:  # No subclass with matching prefix found (& no default defined)
            raise VsaBase.Unknown(
                'ERROR: class "VsaType.{}" is not defined.'.format(vsa_type.name))

    def __array_finalize__(self, obj):
        # see InfoArray.__array_finalize__ for comments
        if obj is None: return
        self.vsa_type = getattr(obj, 'vsa_type', None)
        # if self.vsa_type == VsaType.Laiho:
        self.slots = getattr(obj, 'slots', None)
        self.bits_per_slot = getattr(obj, 'bits_per_slot', None)

    def __reduce__(self):
        # Get the parent's __reduce__ tuple
        pickled_state = super(VsaBase, self).__reduce__()
        # Create our own tuple to pass to __setstate__
        new_state = pickled_state[2] + (self.__dict__,)
        # Return a tuple that replaces the parent's __setstate__ tuple with our own
        return (pickled_state[0], pickled_state[1], new_state)

    def __setstate__(self, state):
        self.__dict__.update(state[-1])  # Update the internal dict from state
        # Call the parent's __setstate__ with the other tuple elements.
        super(VsaBase, self).__setstate__(state[0:-1])


    @classmethod
    def default_numpy_type(cls):
        """
        :return: Should return the default numpy datatype for this type of VSA, e.g., for Vsa.BSC return 'uint8'
        """
        raise NotImplementedError('Subclass must implment property "default_numpy_type()"')

    @classmethod
    def _get_all_subclasses(cls):
        """ Recursive generator of all class' subclasses. """
        for subclass in cls.__subclasses__():
            yield subclass
            for subclass in subclass._get_all_subclasses():
                yield subclass

    @classmethod
    def get_subclass(cls, vsa_type):
        for subclass in cls._get_all_subclasses():
            if subclass.vsatype == vsa_type:
                return subclass
        return None

    @classmethod
    def get_all_converters(cls):
        converters = {}
        for subclass in cls._get_all_subclasses():
            converters[subclass.vsatype] = subclass.get_converters()

        return converters

    @classmethod
    def get_converter(cls, vsa_type):
        raise NotImplementedError('Subclass must implment "validate_operand()"')

    @classmethod
    def validate_operand(cls, b):
        """
        Ensure partner operand is of compatible 'vsatype'
        :param b:
        :return:
        """
        assert b.vsa_type == cls.vsatype, "Mismatch vsa_types"
        return b.vsa_type == cls.vsatype

    @staticmethod
    def trunc_vecs_to_same_len(a, b):
        """
        Because VSA vectors are holgraphic we can compare vectors of different lengths.
        this method ensures that paramters are of the same length by truncating to the shortest of the two.
        :param a: VSA vector
        :param b: VSA vector
        :return: parms (a, b) truncated to match shortest vector.
        """
        if len(a) == len(b):
            return a, b
        elif len(a) > len(b):
            return a[:len(b)], b
        else:
            return a, b[:len(a)]

    @property
    def unpack(self):
        return np.unpackbits(self)

    @property
    def pack(self):
        return VsaBase(np.packbits(self), self.vsa_type)

    @classmethod
    def random_threshold(cls, *args, **kwargs):
        """
        Should return a normalised value of the similarity match that would be expected when comparing
        random/orthorgonal vectors.
        :rtype: float
        """
        raise NotImplementedError('Subclass must implment "random_threshold()"')


    @classmethod
    def randvec(cls, *args, **kwargs):
        """
        :param dims: integer or tuple, specifies shape of required array, last element is no bits per vector.
        :param word_size: numpy's word size parameter, e.g. for BSCs wordsize=8 becomes 'uint8'.
        :param vsa_type: type of VSA subclass to create from VsaType class.
        :return: a matrix of vectors of shape 'dims'.
        """
        raise NotImplementedError('Subclass must implment "randvec()"')

    @classmethod
    def normalize(cls, sv, *args, **kwargs):
        """
        Normalize the VSA vector
        :param a: input VSA vector
        :param seqlength: Optional, needed to normalise BSC vectors.
        :param rv: Optional random vector, used for splitting ties on binary and ternary vectors.
        :return: new VSA vector
        """
        raise NotImplementedError('Subclass must implment "normalize()"')

    @classmethod
    def bind(cls, a, b):
        """
        Comutative binding operator
        :param a: VSA vec
        :param b: VSA vec
        :return: vector associating/coupling a to b that is dissimilar to both a and b.
                 In most cases bind(a, b) is analogues to multiplication, e.g. bind(3,4)=>12.
                 If we know one of the operands we can recover the other using unbind(a,b) e.g unbind(3,12)=>4
        """
        raise NotImplementedError('Subclass must implment "bind()"')

    @classmethod
    def unbind(cls, a, b):
        """
        Comutative unbinding operator. Decouples a from b and vice-versa. The result
        :param a: VSA vec
        :param b: VSA vec
        :return: reverses a bind operation. If z = bind(x, y) then x = unbind(y, z) and y = unbind(x, z).
                 The return is orthogonal to x nd y if x and y have not been previously associated with bind(x, y).
        """
        raise NotImplementedError('Subclass must implment "unbind()"')

    @classmethod
    def hdist(cls, a, b):
        """
        :param a: vsa vector
        :param b: vsa vector
        :return: normalized hamming distance between a and b. 0.0=exact match.
        """
        raise NotImplementedError('Subclass must implment "hdist()"')

    @classmethod
    def hsim(cls, a, b):
        """
        :param a: vsa vector
        :param b: vsa vector
        :return: normalized hamming similarity between a and b. 1.0=exact match.
        """
        raise NotImplementedError('Subclass must implment "hsim()"')


    @classmethod
    def cosine(cls, a, b):
        """
        :param a: vsa vector
        :param b: vsa vector
        :return: cosine distance between a and b, 0.0=exact match.
        """
        if a.vsa_type == VsaType.Tern or a.vsa_type == VsaType.TernZero:
            assert b.vsa_type == VsaType.Tern or b.vsa_type == VsaType.TernZero, "Mismatch vsa_types"
        else:
            assert a.vsa_type == b.vsa_type, "Mismatch vsa_types"
        return sp.distance.cosine(a, b)

    @classmethod
    def cosine_sim(cls, a, b):
        """
        :param a: vsa vector
        :param b: vsa vector
        :return: cosine similarity between a and b. 1.0=exact match.
        """
        return 1.0 - cls.cosine(a, b)

    @classmethod
    def sum(cls, ndarray, *args, **kwargs):
        """
        Maintains vsa_type custom attribute when perfoming numpy.sum()
        Todo: there is probably a better way than this.
        """
        return VsaBase(np.sum(np.array(ndarray), axis=0, *args, **kwargs), vsa_type=ndarray[0].vsa_type)



import numpy as np

from .vsatype import *


class Tern(VsaBase):
    vsatype = VsaType.Tern

    @classmethod
    def default_numpy_type(cls):
        """
        :return: the default numpy datatype for this type of VSA.
        """
        return 'int8'

    @classmethod
    def unpackbits(cls, v):
        vbin = np.unpackbits(v).astype('int8')
        vbin[vbin == 0] = -1
        return VsaBase(vbin, vsa_type=VsaType.Tern)

    @classmethod
    def packbits(cls, v):
        v[v == -1] = 0
        return np.packbits(v)

    @classmethod
    def random_threshold(cls, *args, stdev_count=4.4, **kwargs):
        from .bsc import BSC
        return BSC.random_threshold(*args, stdev_count=stdev_count, **kwargs)

    @classmethod
    def randvec(cls, dims, word_size=8, vsa_type=VsaType.Tern):
        """
        :param dims: integer or tuple, specifies shape of required array, last element is no bits per vector.
        :param word_size: numpy's word size parameter, e.g. for BSCs wordsize=8 becomes 'uint8'.
        :param vsa_type: type of VSA subclass to create from VsaType class.
        :return: a matrix of vectors of shape 'dims'.
        """
        v = np.random.randint(0, 2, size=dims, dtype='int' + str(word_size))
        v[v == 0] = -1
        return VsaBase(v, vsa_type=vsa_type)

    @classmethod
    def bind(cls, a, b):  # actually bind/unbind for binary and ternary vecs
        """
        Comutative binding operator
        :param a: VSA vec
        :param b: VSA vec
        :return: vector associating/coupling a to b that is dissimilar to both a and b.
                 In most cases bind(a, b) is analogues to multiplication, e.g. bind(3,4)=>12.
                 If we know one of the operands we can recover the other using unbind(a,b) e.g unbind(3,12)=>4
        """
        assert a.vsa_type == VsaType.Tern or b.vsa_type == VsaType.TernZero, "Mismatch vsa_types"
        if b.vsa_type == VsaType.Tern or b.vsa_type == VsaType.TernZero:
            return -(a * b)
        raise ValueError

    @classmethod
    def unbind(cls, a, b):
        """
        Comutative unbinding operator. Decouples a from b and vice-versa. The result
        :param a: VSA vec
        :param b: VSA vec
        :return: reverses a bind operation. If z = bind(x, y) then x = unbind(y, z) and y = unbind(x, z).
                 The return is orthogonal to x nd y if x and y have not been previously associated with bind(x, y).
        """
        return cls.bind(a, b)

    @classmethod
    def normalize(cls, sv, seqlength=None, rv=None):
        """
        Normalize the VSA vector
        :param a: input VSA vector
        :param seqlength: not used
        :param rv: Optional random vector, used for splitting ties on binary and ternary VSA vectors.
        :return: new VSA vector
        """
        v = sv.copy()  # we don't want to change sv
        if rv is None:
            rv = cls.randvec(len(v), vsa_type=sv.vsa_type)
        v[v == 0] = rv[v == 0]  # Randomise the zeros to +1 or -1
        v[v < 0] = -1.0  # all elements < 0 become -1
        v[v > 0] = 1.0  # all elements > 0 become +1
        return v

    @classmethod
    def hdist(cls, a, b):
        """
        :param a: vsa vector
        :param b: vsa vector
        :return: normalized hamming distance between a and b. 0.0=exact match.
        """
        return 1.0 - cls.hsim(a,b)

    @classmethod
    def hsim(cls, a, b):
        """
        :param a: vsa vector
        :param b: vsa vector
        :return: normalized hamming distance between a and b. 1.0=exact match.
        """
        assert b.vsa_type == VsaType.Tern or b.vsa_type == VsaType.TernZero, "Mismatch vsa_types"
        if b.vsa_type == VsaType.Tern or b.vsa_type == VsaType.TernZero:
            x = cls.bind(a, b) * -1

            if np.count_nonzero(x) == len(x):
                num_ones = (x == 1)  # Gets an array of the +1's only (will be shorter than the full length)
                return np.count_nonzero(num_ones) / len(x)
            else:
                # For TernZero class, if there are zeros allowed in the array we can simulate the HDSim like below.
                # Doing it this way means we keep the HDSim scale relative to BSC and Tern
                num_zeros = np.count_nonzero(x == 0)  # Count zeros in the array
                hd1 = np.count_nonzero(x == 1)  # Count 1's in the array
                hd1 = (hd1 + num_zeros * 0.5) / len(x)  # Add half of the don't care bits (zeros) to teh HDSim

                #################################################################################################
                # Note the alternate way shown below changes the TernZero HDSim scale relative to Tern and BSC
                # making the HDSim value relatively higher whereas the method above matches the relative scales.
                #    num_minus1s = np.count_nonzero(x == -1)  # Count -1's in the array
                #    num_ones = np.count_nonzero(x == 1)  # Count 1's in the array
                #    hd1 = num_ones / (num_ones+num_minus1s)  # The 'active vector length' = (num_ones + num_minus1s)
            return hd1
        raise ValueError




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vsapy']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.1.1,<10.0.0',
 'gmpy2>=2.1.2,<3.0.0',
 'nltk>=3.7,<4.0',
 'numpy>=1.22.4,<2.0.0',
 'pandas>=1.4.2,<2.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'regex>=2022.4.24,<2023.0.0',
 'scipy>=1.8.1,<2.0.0',
 'setuptools>=62.3.2,<63.0.0',
 'six>=1.16.0,<2.0.0',
 'unicodedata2>=14.0.0,<15.0.0',
 'wheel>=0.37.1,<0.38.0',
 'xmltodict>=0.13.0,<0.14.0']

setup_kwargs = {
    'name': 'vsapy',
    'version': '0.7.0',
    'description': 'Vector Symbolic Architecture(VSA) library that allows building VSA apps that use various flavours of VSA vectors.',
    'long_description': None,
    'author': 'Chris Simpkin',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/vsapy/vsapy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
